; Author: Juergen Lorenz ; ju (at) jugilo (dot) de
;
; Copyright (c) 2013-2017, Juergen Lorenz
; All rights reserved.
; 
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are
; met:
; 
; Redistributions of source code must retain the above copyright
; notice, this list of conditions and the following dispasser.
; 
; Redistributions in binary form must reproduce the above copyright
; notice, this list of conditions and the following dispasser in the
; documentation and/or other materials provided with the distribution.
; 
; Neither the name of the author nor the names of its contributors may be
; used to endorse or promote products derived from this software without
; specific prior written permission. 
; 
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
; IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
; TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
; PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
; HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
; TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
; PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
; LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
; NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#|[
This library will provide some macro-writing macros, in particular
macro-rules and define-macro, based on explicit- and implicit-renaming.
The syntax of macro-rules mimics that of syntax-rules, except that it
allows for injected symbols before the keyword list and the templates
are usually quasiquoted lists. Since we use bind-case from the bindings
egg, this library accepts wildcards, non-symbol literals and fenders.
]|#

(require-library basic-macros)

(module procedural-macros
  (define-macro macro-rules macro-let macro-letrec)
  
  (import scheme
          (only basic-macros bind-case define-er-macro-transformer)
          (only chicken print error case-lambda))
  (import-for-syntax (only basic-macro-helpers filter pseudo-flatten))

  (reexport (only basic-macros
                  once-only
                  define-ir-macro-transformer
                  define-er-macro-transformer
                  define-ir-macro
                  define-er-macro
                  with-mapped-symbols
                  with-gensyms
                  ))

#|[
The workhorse of the library is the following macro, a procedural
version of syntax-rules, but without its limitations.
]|#

;;; (macro-rules sym ... (key ...) (pat tpl) ...)
;;; (macro-rules sym ... (key ...) (pat (where fender ...) tpl) ...)
;;; ----------------------------------------------------------------
;;; where sym ... are injected non-hygienig symbols, key ... are
;;; additional keywords, pat ....  are nested lambda-lists without
;;; spezial meaning of ellipses and tpl .... usually evaluate to
;;; quasiquoted templates. The optional fenders belong to the pattern
;;; matching process.
(define-er-macro-transformer (macro-rules f r c?)
  (let (
    (f* (let loop ((tail (cdr f)) (head '()))
          (if (symbol? (car tail))
            (loop (cdr tail) (cons (car tail) head))
            (cons head tail))))
    (%x (r 'x))
    (%let (r 'let))
    (%form (r 'form))
    (%where (r 'where))
    (%lambda (r 'lambda))
    (%inject (r 'inject))
    (%compare? (r 'compare?))
    (%bind-case (r 'bind-case))
    (%ir-macro-transformer (r 'ir-macro-transformer))
    )
    (let ((syms (car f*))
          (keys (cadr f*))
          (rules (cddr f*)))
      (let* ((pats (map car rules))
             (fpats (map pseudo-flatten pats))
             (kpats (map (lambda (fp)
                           (filter (lambda (x)
                                     (memq x keys))
                                   fp))
                         fpats))
             ;; compare? keywords with its names
             (key-checks
               (map (lambda (kp)
                      (map (lambda (p s)
                             `(,p (,%lambda (,%x)
                                            (,%compare? ,%x ,s))))
                           kp
                           (map (lambda (x) `',x)
                                kp)))
                    kpats))
             ;; prepare where clause for each rule
             ;; to check keys
             (all-rules (map (lambda (rule checks)
                               (let ((second (cadr rule)))
                                 (if (and (pair? second)
                                          (c? (car second) %where))
                                   `(,(car rule)
                                      (,%where ,@(cdr second) ,@checks)
                                      ,@(cddr rule))
                                   `(,(car rule)
                                      (,%where ,@checks)
                                      ,@(cdr rule)))))
                             rules key-checks)))
        `(,%ir-macro-transformer
           (,%lambda (,%form ,%inject ,%compare?)
             (,%let ,(map (lambda (s)
                       `(,s (,%inject ',s)))
                     syms)
               (,%bind-case ,%form
                            ,@all-rules))))))))

#|[
And now a procedural version of our old friend, define-macro,
which is hygienic, if now injections are provided.
]|#

;;; (define-macro (name . args)
;;;   (where (x . xs) ...)
;;;   xpr . xprs)
;;; -----------------------------------  
;;; where xs is either a list of predicates, thus providing fenders,
;;; or a singleton containing one of the symbols keyword or injection
;;; to provide keyword arguments or nonhygienic macros 
(define-er-macro-transformer (define-macro form rename compare?)
  (let ((code (cadr form))
        (xpr (caddr form))
        (xprs (cdddr form))
        (%where (rename 'where))
        (%keyword (rename 'keyword))
        (%injection (rename 'injection))
        (%define-macro (rename 'define-macros))
        (%macro-rules (rename 'macro-rules))
        (%define-syntax (rename 'define-syntax)))
    (let ((name (car code)) (args (cdr code)))
      (if (and (pair? xpr)
               (compare? (car xpr) %where)
               (not (null? xprs)))
        (let ((clauses (cdr xpr)))
          (let (
            (fenders
              (filter (lambda (clause)
                        (or (null? (cdr clause))
                            (and (not (compare? (cadr clause) %keyword))
                                 (not (compare? (cadr clause) %injection)))))
                      clauses))
            (keywords
              (filter (lambda (clause)
                        (and (not (null? (cdr clause)))
                             (compare? (cadr clause) %keyword)))
                      clauses))
            (injections
              (filter (lambda (clause)
                        (and (not (null? (cdr clause)))
                             (compare? (cadr clause) %injection)))
                      clauses))
            )
            (let (
              (keywords
                (if (null? keywords)
                  keywords
                  (map car keywords)))
              (injections
                (if (null? injections)
                  injections
                  (map car injections)))
              )
              `(,%define-syntax ,name
                 (,%macro-rules ,@injections ,keywords
                   ((_ ,@args) (where ,@fenders) ,@xprs))))))
        `(,%define-syntax ,name
           (,%macro-rules ()
             ((_ ,@args) ,xpr ,@xprs)))))))

#|[
Now follow the local versions of define-macro, macro-let and
macro-letrec. Since the syntax of both is identical, they are
implemented by means of a helper macro.
]|#

;; helper for macro-let and macro-letrec
(define-er-macro-transformer (macro-with form rename compare?)
  (let ((op (cadr form))
        (pat-tpl-pairs (caddr form))
        (xpr (cadddr form))
        (xprs (cddddr form))
        (%macro-rules (rename 'macro-rules)))
    (let ((pats (map car pat-tpl-pairs))
          (tpls (map cdr pat-tpl-pairs)))
      `(,op ,(map (lambda (pat tpl)
                    `(,(car pat)
                       (,%macro-rules ()
                         ((_ ,@(cdr pat)) ,@tpl))))
                  pats tpls)
                   ,xpr ,@xprs))))

;;; (macro-let (((name . args) (where fender ...) .. xpr ...) ...) body ....)
;;; -------------------------------------------------------------------------
;;; evaluates body ... in the context of parallel macros name ....
(define-er-macro-transformer (macro-let form rename compare?)
  (let ((pat-tpl-pairs (cadr form))
        (xpr (caddr form))
        (xprs (cdddr form));)
        (%macro-with (rename 'macro-with))
        (%let-syntax (rename 'let-syntax)))
    `(,%macro-with ,%let-syntax ,pat-tpl-pairs ,xpr ,@xprs)))

;;; (macro-letrec (((name . args) (where fender ...) .. xpr ...) ...) body ....)
;;; ----------------------------------------------------------------------------
;;; evaluates body ... in the context of recursive macros name ....
(define-er-macro-transformer (macro-letrec form rename compare?)
  (let ((pat-tpl-pairs (cadr form))
        (xpr (caddr form))
        (xprs (cdddr form));)
        (%macro-with (rename 'macro-with))
        (%letrec-syntax (rename 'letrec-syntax)))
    `(,%macro-with ,%letrec-syntax ,pat-tpl-pairs ,xpr ,@xprs)))

;;; (procedural-macros sym ..)
;;; --------------------------
;;; documentation procedure.
(define procedural-macros
  (let ((alst '(
    (macro-rules
      macro:
      (macro-rules literal ... (keyword ...) (pat (where fender ...) .. tpl) ....)
      "procedural version of syntax-rules"
      "with optional injected literals"
      "and quasiquoted templates")
    (define-macro
      macro:
      (define-macro (name . args) (where (x . xs) ...) .. xpr ....)
      "a version of macro-rules with only one rule"
      "xs is either a list of predicates, thus providing fenders"
      "or a singleton containing one of the symbols keyword or"
      "injection, providing keyword parameters or nonhygienic macros")
    (macro-let
      macro:
      (macro-let (((name . args) (where fender ...) .. xpr ...) ...) body ....)
      "evaluates body ... in the context of parallel macros name ....")
    (macro-letrec
      macro:
      (macro-letrec (((name . args) (where fender ...) .. xpr ...) ...) body ....)
      "evaluates body ... in the context of recursive macros name ....")
    (once-only
      macro:
      (once-only (x ....) xpr ....)
      "arguments x ... are evaluated only once and"
      "from left to right in the body xpr ....")
    (define-er-macro-transformer
      macro:
      (define-er-macro-tansformer name form rename compare?)
      "wrapper around er-macro-transformer")
    (define-ir-macro-transformer
      macro:
      (define-ir-macro-tansformer name form inject compare?)
      "wrapper around ir-macro-transformer")
    (define-er-macro
      macro:
      (define-er-macro name form rename-prefix compare?)
      "creates an explicit-renaming macro, where all symbols"
      "starting with rename-prefix are renamed automatically")
    (define-ir-macro
      macro:
      (define-ir-macro name form inject-prefix compare?)
      "creates an implicit-renaming macro, where all symbols"
      "starting with inject-prefix are injected automatically")
    (with-mapped-symbols
      macro:
      (with-mapped-symbols mapper prefix- (prefix-x ....) xpr ....)
      "binds a series of prefixed names, prefix-x ...."
      "to the images of the original names, x ...., under mapper"
      "and evaluates xpr .... in this context")
    (with-gensyms
      macro:
      (with-gensyms (x ....) xpr ....)
      "generates a series of gensyms x .... to be used in body xpr ...")
    )))
    (case-lambda
      (()
       (map car alst))
      ((sym)
       (let ((lst (assq sym alst)))
         (if lst
           (for-each print (cdr lst))
           (error 'procedural-macros
                  "not exported" sym)))))))
) ; procedural-macros

